/*	
 *	cliprend.h
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001 Naoyuki Sawa
 *
 *	* Wed Feb 13 06:00:00 JST 2001 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "clip.h"

/*****************************************************************************
 *	
 *****************************************************************************/

TEXTURE*
texture_init(TEXTURE* texture, const void* data)
{
	TEXTUREHEADER* header;

	header = (TEXTUREHEADER*)data;

	/* wb_܂B
	 * * P/ECE̎dlɁu1bitf[^masktv(bpp=1,mask=1)݂͑܂񂪁A
	 *   {Cuł́ALȃeNX``ƂĂƂėp\Ȃ̂ŁA
	 *   G[ɂ͂܂B
	 * * eNZWƂĎw\Ȃ̂0`255ɐ܂A
	 *   eNX`̂̃TCY256hbgȏł\܂B
	 *   ]āAł̓TCY̏͌Ă܂B
	 */
	if(strncmp((char*)&header->head, "PMBP", 4) != 0) DIE();
	if(header->bpp  != 1 && header->bpp  != 2) DIE();
	if(header->mask != 0 && header->mask != 1) DIE();
	if(header->w < 8 || (header->w & 7) != 0) DIE();
	if(header->h < 8 || (header->h & 7) != 0) DIE();

	/* eNX`wb_ƃ|C^i[܂B */
	memset(texture, 0, sizeof(TEXTURE));
	texture->header = *header;
	texture->buf  = (unsigned char*)(header + 1);
	if(header->mask) {
		texture->mask = (unsigned char*)(header + 1) +
				header->w * header->h / (8 / header->bpp);
	}

	return texture;
}

/*****************************************************************************
 *	RENDER
 *****************************************************************************/

#define N_VERTICES_MAX	8	/* 1|S̍ő咸_ */

/* NbsOp_ */
typedef struct _CLIPVERTEX {
	float x, y, z;		/* _W */
	float s, t;		/* eNX`W */
} CLIPVERTEX;

/* `p_ */
typedef struct _RENDVERTEX {
	unsigned char x, y;	/* _W */
	unsigned short z;	/* yl(8.8bit) */
	unsigned short s, t;	/* eNX`W(8.8bit) */
} RENDVERTEX;

/* DDA */
typedef struct _DDA {
	int n;			/* [̑     */
	int d;			/* [             */
	int u;			/* ̕ω     */
	int v;			/* Jグ̕ω */
} DDA;

static CLIPVERTEX clip_vertices[2][N_VERTICES_MAX];	/* NbsOp_obt@ */
static RENDVERTEX rend_vertices[N_VERTICES_MAX];	/* `p_obt@         */
static unsigned char fore_color;			/* OiF                     */
static unsigned char back_color;			/* wiF                     */

static int  z_clip(RENDER* render, int nsrc);
static int xy_clip(RENDER* render, int nsrc);
static void draw_primitive(RENDER* render, int n_vertices);

/* OPT_DRAW_SCANLINE`ƁAXLC`R[hWJ܂B(xAbvATCY)
 * `RgAEgƁAXLC`R[hWJ܂B (x_EATCY)
 * OPT_DRAW_SCANLINE`ꍇAgCAOZbgAbvł̓eNX`WvZ܂B
 * ͑SʂȏłAR[hTCYƂ̌ˍŁAdraw_scanline()WJ邱Ƃɂ܂B
 */
#define OPT_DRAW_SCANLINE
#ifndef OPT_DRAW_SCANLINE
static void draw_scanline(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
#else /*OPT_DRAW_SCANLINE*/
static void (*draw_scanline)(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_color(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_1bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_1bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_2bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
static void draw_scanline_2bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1);
#endif /*OPT_DRAW_SCANLINE*/

void
render_init(RENDER* render, SURFACE* surface, int x, int y, int width, int height, float front, float back, float fovy)
{
	/* T[tFCXݒ肵܂B */
	render->surface = surface;

	/* r[|[g`ݒ肵܂B */
	if(width == 0) width = surface->w;
	if(height == 0) height = surface->h;
	render->left = x;
	render->top = y;
	render->right = x + width;
	render->bottom = y + height;
	if(render->left < 0 ) render->left = 0;
	if(render->top < 0 ) render->top = 0;
	if(render->right > surface->w) render->right = surface->w;
	if(render->bottom > surface->h) render->bottom = surface->h;

	/* NbsOݒ肵܂B */
	render->front = front;
	render->back = back;

	/* ReLXgX^bN܂B */
	render->context = render->context_stack;

	/* ϊs܂B */
	matrix_viewportz(CURRENT_MATRIX(render), &MATRIX_I, x, y, width, height);
	matrix_perspect (CURRENT_MATRIX(render), CURRENT_MATRIX(render), fovy, (float)width / height, front);

	/* eNX`܂B */
	CURRENT_TEXTURE(render) = NULL;
}

void
render_save(RENDER* render)
{
	/* X^bNgbv𕡐܂B */
	if(++render->context == render->context_stack + RENDER_CONTEXT_STACK_SIZE) DIE();
	memcpy(render->context, render->context - 1, sizeof(RENDERCONTEXT));
}

void
render_restore(RENDER* render)
{
	/* X^bNgbv̂Ă܂B */
	if(render->context-- == render->context_stack) DIE();
}

void
render_line(RENDER* render, int x0, int y0, int x1, int y1, int color)
{
	/* NbsOp}NB */
#define CLIP(X, Y, A, B, OP, LIM) \
	if(!((X##A) OP (LIM))) { \
		if(!((X##B) OP (LIM))) { \
			return; \
		} else { \
			(Y##A) = (Y##A) + \
			((Y##B) - (Y##A)) * ((LIM) - (X##A)) / ((X##B) - (X##A)); \
			(X##A) = (LIM); \
		} \
	}

	SURFACE* surface;
	unsigned char* vptr;
	int nx, ny;
	int ux, uy;
	int d, n;
	int tmp, ofs;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* e_AE㉺ŃNbv܂B */
	CLIP(x, y, 0, 1, >=, render->left);
	CLIP(x, y, 0, 1, <=, render->right - 1);
	CLIP(x, y, 1, 0, >=, render->left);
	CLIP(x, y, 1, 0, <=, render->right - 1);
	CLIP(y, x, 0, 1, >=, render->top);
	CLIP(y, x, 0, 1, <=, render->bottom - 1);
	CLIP(y, x, 1, 0, >=, render->top);
	CLIP(y, x, 1, 0, <=, render->bottom - 1);

	/* n_AhX߂܂B */
	ofs = surface->w * y0 + x0;
	vptr = surface->vbuff + ofs;

	/* XDDAp^vZ܂B */
	if(x1 >= x0) {
		nx = x1 - x0;
		ux =  1;
	} else {
		nx = x0 - x1;
		ux = -1;
	}

	/* YDDAp^vZ܂B */
	if(y1 >= y0) {
		ny = y1 - y0;
		uy =  surface->w;
	} else {
		ny = y0 - y1;
		uy = -surface->w;
	}

	/* 厲(X)A𕛎(Y)Ƃ܂B */
	if(nx < ny) {
		tmp = x0; x0 = y0; y0 = tmp;
		tmp = x1; x1 = y1; y1 = tmp;
		tmp = nx; nx = ny; ny = tmp;
		tmp = ux; ux = uy; uy = tmp;
	}
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDA[vB */
	d = -nx;
	n = nx;
	do {
		*vptr = color;
		vptr += ux;
		d += ny * 2;
		if(d >= 0) {
			vptr += uy;
			d -= nx * 2;
		}
	} while(--n); /* I_͕`Ȃ̂ŁAvfNgłB */

#undef CLIP
}

void
render_mesh(RENDER* render, const MESH* mesh)
{

	/* ϊςݒ_Wobt@B
	 * MESH.nv8rbgȂ̂ŁAő255_łB
	 * P/ECE̓X^bN̂ŁAstaticɔzu܂B
	 */
	static VECTOR xformed_vectors[255];

	int i, j, n_vectors, n_coords, n_faces, n_face_vertices, n_vertices;
	float s, xN0, yN0, x01, y01;
	VECTOR *vectors, *vector, *vector_src, *vector_dst;
	COORD *coords, *coord;
	FACE *face;
	FACEVERTEX *face_vertices, *face_vertex;
	CLIPVERTEX *clip_vertex;
	RENDVERTEX *rend_vertex;

#ifdef OPT_DRAW_SCANLINE
	/* XLC`R[hI܂B */
	TEXTURE* texture;
	texture = CURRENT_TEXTURE(render);
	if(texture == NULL) {
		draw_scanline = draw_scanline_color;
	} else {
		if(!texture->header.mask) {
			switch(texture->header.bpp) {
			case 1:
				draw_scanline = draw_scanline_1bit;
				break;
			default/*2*/:
				draw_scanline = draw_scanline_2bit;
				break;
			}
		} else {
			switch(texture->header.bpp) {
			case 1:
				draw_scanline = draw_scanline_1bitmask;
				break;
			default/*2*/:
				draw_scanline = draw_scanline_2bitmask;
				break;
			}
		}
	}
#endif /*OPT_DRAW_SCANLINE*/

	/* Xg̐擪AhXvZ܂B */
	n_vectors = mesh->nv;
	n_coords = mesh->nc;
	n_faces = mesh->nf;
	vectors = (VECTOR*)(mesh + 1);
	coords = (COORD*)(vectors + n_vectors);
	face = (FACE*)(coords + n_coords);

	/* S_Wϊ܂B */
	for(i = 0, vector_src = vectors, vector_dst = xformed_vectors;
	    i < n_vectors;
	    i++, vector_src++, vector_dst++) {
		vector_xform(vector_dst, CURRENT_MATRIX(render), vector_src);
	}

	/* etFCX`܂B */
	for(i = 0;
	    i < n_faces;
	    i++, face = (FACE*)(face_vertices + n_face_vertices)) {
		n_face_vertices = face->nfv;
		fore_color = face->c >> 4 & 15;
		back_color = face->c      & 15;
		face_vertices = (FACEVERTEX*)(face + 1);

		/* NbsOp_XggݗĂ܂B
		 * * eNX`̏ꍇsĂ܂񂪁AɖȂ͂łB
		 *   eNX`Ȃmesh->nc=0Ȃ̂ŁAcoords͍ŏ̃tFCXwĂ܂B
		 *   etFCX_face_vertex->ic=0Ȃ̂ŁAeNX`WƂčŏ̃tFCXm,n擾܂B
		 *   mesh->nf=0̏ꍇ̂ݍŏ̃tFCX݂܂񂪁Ȁꍇ͂̃[v̂܂B
		 *   擾eNX`W͕`掞ɖ̂ŁAeNX`Wl͊֌W܂B
		 *   L̒ʂAeNX`W擾ɕsȃ̈ɃANZXȂƂۏ؂܂B
		 *   ]āAeNX`̏ꍇsȂĂAɖ͂ȂƎv܂B
		 */
		n_vertices = 0;
		for(j = 0, face_vertex = face_vertices, clip_vertex = clip_vertices[0];
		    j < n_face_vertices;
		    j++, face_vertex++, clip_vertex++) {
			vector = &xformed_vectors[face_vertex->iv];
			coord = &coords[face_vertex->ic];
			clip_vertex->x = vector->x;
			clip_vertex->y = vector->y;
			clip_vertex->z = vector->z;
			clip_vertex->s = coord->s;
			clip_vertex->t = coord->t;
			if(++n_vertices == N_VERTICES_MAX) break;
		}
		if(n_vertices < 3) continue;

		/* yNbsOs܂B */
		n_vertices = z_clip(render, n_vertices);
		if(n_vertices < 3) continue;

		/* ϊ̂߂ x=x/z,y=y/z s܂B */
		for(j = 0, clip_vertex = clip_vertices[0];
		    j < n_vertices;
		    j++, clip_vertex++) {
			clip_vertex->x /= clip_vertex->z;
			clip_vertex->y /= clip_vertex->z;
		}

		/* \s܂B
		 * E(=\)LqȂ̂ŁA(=)Ȃ`܂B
		 */
		xN0 = clip_vertices[0][0].x - clip_vertices[0][n_vertices - 1].x;
		yN0 = clip_vertices[0][0].y - clip_vertices[0][n_vertices - 1].y;
		x01 = clip_vertices[0][1].x - clip_vertices[0][             0].x;
		y01 = clip_vertices[0][1].y - clip_vertices[0][             0].y;
		s = xN0 * y01 - x01 * yN0;
		if(s <= 0) continue;

		/* ʃNbsOs܂B */
		n_vertices = xy_clip(render, n_vertices);
		if(n_vertices < 3) continue;

		/* `p_XggݗĂ܂B */
		for(j = 0, clip_vertex = clip_vertices[0], rend_vertex = rend_vertices;
		    j < n_vertices;
		    j++, clip_vertex++, rend_vertex++) {
			rend_vertex->x = (unsigned char)clip_vertex->x;
			rend_vertex->y = (unsigned char)clip_vertex->y;
			rend_vertex->z = (unsigned short)((clip_vertex->z - render->front) * render->back   /
							  (render->back   - render->front) / clip_vertex->z * 0xFFFF);
			rend_vertex->s = (unsigned short)(clip_vertex->s * 0x100);
			rend_vertex->t = (unsigned short)(clip_vertex->t * 0x100);
		}

		/* v~eBu`܂B */
		draw_primitive(render, n_vertices);
	}
}

static int
z_clip(RENDER* render, int nsrc)
{
#define CLIP(SRC, DST, OP, LIM) \
	src = SRC; \
	dst = DST; \
	ndst = 0; \
	lim = LIM; \
	for(i = 0; i < nsrc; i++) { \
		v0 = src +  i; \
		v1 = src + (i + 1) % nsrc; \
		if(v0->z OP lim) { \
			*dst = *v0; \
			dst++; \
			ndst++; \
			if(ndst == N_VERTICES_MAX) break; \
			if(v1->z OP lim) { \
				continue; \
			} else { \
				/* go clip */ \
			} \
		} else { \
			if(v1->z OP lim) { \
				/* go clip */ \
			} else { \
				continue; \
			} \
		} \
		/* do clip */ \
		dst->z = lim; \
		n0 = lim - v0->z; \
		n1 = v1->z - lim; \
		n01 = n0 + n1; \
		dst->x = v0->x + (v1->x - v0->x) * n0 / n01; \
		dst->y = v0->y + (v1->y - v0->y) * n0 / n01; \
		dst->s = v0->s + (v1->s - v0->s) * n0 / n01; \
		dst->t = v0->t + (v1->t - v0->t) * n0 / n01; \
		dst++; \
		ndst++; \
		if(ndst == N_VERTICES_MAX) break; \
	} \
	if(ndst < 3) return 0; \

	CLIPVERTEX *src, *dst;
	CLIPVERTEX *v0, *v1;
	float lim, n0, n1, n01;
	int i, ndst;

	CLIP(clip_vertices[0], clip_vertices[1], >=, render->front); nsrc = ndst;
	CLIP(clip_vertices[1], clip_vertices[0], <=, render->back );
	return ndst;

#undef CLIP
}

int
xy_clip(RENDER* render, int nsrc)
{
#define CLIP(SRC, DST, X, Y, OP, LIM) \
	src = SRC; \
	dst = DST; \
	ndst = 0; \
	lim = LIM; \
	for(i = 0; i < nsrc; i++) { \
		v0 = src +  i; \
		v1 = src + (i + 1) % nsrc; \
		if(v0->X OP lim) { \
			*dst = *v0; \
			dst++; \
			ndst++; \
			if(ndst == N_VERTICES_MAX) break; \
			if(v1->X OP lim) { \
				continue; \
			} else { \
				/* go clip */ \
			} \
		} else { \
			if(v1->X OP lim) { \
				/* go clip */ \
			} else { \
				continue; \
			} \
		} \
		/* do clip */ \
		dst->X = lim; \
		n0 = lim - v0->X; \
		n1 = v1->X - lim; \
		n01 = n0 + n1; \
		dst->Y = v0->Y + (v1->Y - v0->Y) * n0 / n01; \
		n0 *= v0->z; \
		n1 *= v1->z; \
		n01 = n0 + n1; \
		dst->z = v0->z + (v1->z - v0->z) * n0 / n01; \
		dst->s = v0->s + (v1->s - v0->s) * n0 / n01; \
		dst->t = v0->t + (v1->t - v0->t) * n0 / n01; \
		dst++; \
		ndst++; \
		if(ndst == N_VERTICES_MAX) break; \
	} \
	if(ndst < 3) return 0; \

	CLIPVERTEX *src, *dst;
	CLIPVERTEX *v0, *v1;
	float lim, n0, n1, n01;
	int i, ndst;

	CLIP(clip_vertices[0], clip_vertices[1], x, y, >=, render->left  ); nsrc = ndst;
	CLIP(clip_vertices[1], clip_vertices[0], x, y, <=, render->right ); nsrc = ndst;
	CLIP(clip_vertices[0], clip_vertices[1], y, x, >=, render->top   ); nsrc = ndst;
	CLIP(clip_vertices[1], clip_vertices[0], y, x, <=, render->bottom);
	return ndst;

#undef CLIP
}

static void
draw_primitive(RENDER* render, int n_vertices)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	side->cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		side->i.n = (v1->i - v0->i) % side->ny; \
		side->i.d = -side->ny; \
		side->i.u =  (v1->i - v0->i) / side->ny; \
		side->i.v =  1; \
	} else { \
		side->i.n = (v0->i - v1->i) % side->ny; \
		side->i.d = -side->ny; \
		side->i.u = -(v0->i - v1->i) / side->ny; \
		side->i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	side->cur.i += side->i.u; \
	side->i.d += side->i.n * 2; \
	if(side->i.d >= 0) { \
		side->cur.i += side->i.v; \
		side->i.d -= side->ny * 2; \
	}

	struct {
		int iv;		/* ݂̕ӂ̏I_CfNX            */
		int ny;		/* ݂̕ӂ̍̕ψ                */
		int n;		/* ݂̕ӂ̎c胉C              */
		RENDVERTEX cur;	/* ݒl                              */
		DDA x, z;	/* _WDDA(y͍Eʂ1Âi) */
		DDA s, t;	/* eNX`WDDA                   */
	} sides[2], *side;

	int i, itop, y, top, bottom;
	RENDVERTEX *v, *v0, *v1;

	/* ㉺[̈ʒuƁA[̒_߂܂B */
	itop = 0;
	top = bottom = rend_vertices[0].y;
	for(i = 1; i < n_vertices; i++) {
		v = &rend_vertices[i];
		if(v->y < top) {
			itop = i;
			top = v->y;
		}
		if(v->y > bottom) {
			bottom = v->y;
		}
	}
	if(top == bottom) return; /* |S͕`܂B */

	/* E̕ӂ܂B */
	for(i = 0; i < 2; i++) {
		side = &sides[i];
		side->iv = itop;
		side->n = 0; /* ɖ{̏s悤 */
	}

	/* [AeXLCɂ... */
	for(y = top; y < bottom; y++) {
		/* E̕ӂɂ... */
		for(i = 0; i < 2; i++) {
			side = &sides[i];
			/* ̕ӂɒB... */
			if(side->n == 0) {
				do {
					/* ̕ӂ̗[̒_߂܂B */
					v0 = &rend_vertices[side->iv];
					if(i == 0) {
						side->iv--;
						if(side->iv == -1) {
							side->iv = n_vertices - 1;
						}
					} else {
						side->iv++;
						if(side->iv == n_vertices) {
							side->iv = 0;
						}
					}
					v1 = &rend_vertices[side->iv];
					/* ̕ӂ̍̕ψʂ߂܂B */
					side->ny = v1->y - v0->y;
					/* V|SoAc̕`̓LZ܂B */
					if(side->ny < 0) return;
				} while(side->ny == 0); /* ȕӂ͖܂B */
				/* DDAZbgAbvB */
				SETUP(x);
				SETUP(z);
				SETUP(s);
				SETUP(t);
				/* ̕ӂ܂ł̃JE^B */
				side->n = side->ny;
			}
		}
		/* E̕ӂ̊ԂɁA`܂B */
		sides[0].cur.y = sides[1].cur.y = y;
		draw_scanline(render, &sides[0].cur, &sides[1].cur);
		/* E̕ӂɂ... */
		for(i = 0; i < 2; i++) {
			side = &sides[i];
			/* ̕ӂ܂ł̃C炵܂B */
			side->n--;
			/* DDAXebvsB */
			STEP(x)
			STEP(z)
			STEP(s)
			STEP(t)
		}
	}
	/* [̃XLC͕`܂B */

#undef SETUP
#undef STEP
}

#ifndef OPT_DRAW_SCANLINE
static void
draw_scanline(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface =  render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = CURRENT_TEXTURE(render);

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			if(texture == NULL) {
				/* eNX`ȂAtFCX̔wiFŕ`܂B */
				*vptr = back_color;
				*zptr = zval;
			} else {
				/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
				 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
				 * R[hTCYƂ̌ˍŁÂőË܂B
				 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
				 */
				sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
				tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
				ofs = texture->header.w * tval + sval;
				if(!texture->header.mask || texture->mask[ofs / 8] >> (7 - (ofs & 7)) & 1) {
					switch(texture->header.bpp) {
					case 1:
						/* 1biteNX`ȂAeNZlɂ 1:tFCX̑OiF/0:tFCX̔wiF ŕ`܂B */
						*vptr = texture->buf[ofs / 8] >> (7 - (ofs & 7)) & 1 ? fore_color : back_color;
						break;
					default/*2*/:
						/* 2biteNX`ȂAeNZl̂܂ܕ`܂B */
						*vptr = texture->buf[ofs / 4] >> (3 - (ofs & 3)) * 2 & 3;
						break;
					}
					*zptr = zval;
				}
			}
		}
		/* J[Eyobt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}
#else /*OPT_DRAW_SCANLINE*/
static void
draw_scanline_color(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* eNX`ȂAtFCX̔wiFŕ`܂B */
			*vptr = back_color;
			*zptr = zval;
		}
		/* J[Eyobt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}
static void
draw_scanline_1bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = CURRENT_TEXTURE(render);

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
			 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
			 * R[hTCYƂ̌ˍŁÂőË܂B
			 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
			 */
			sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
			tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
			ofs = texture->header.w * tval + sval;
			/* 1biteNX`ȂAeNZlɂ 1:tFCX̑OiF/0:tFCX̔wiF ŕ`܂B */
			*vptr = texture->buf[ofs / 8] >> (7 - (ofs & 7)) & 1 ? fore_color : back_color;
			*zptr = zval;
		}
		/* J[Eyobt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}
static void
draw_scanline_1bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = CURRENT_TEXTURE(render);

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
			 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
			 * R[hTCYƂ̌ˍŁÂőË܂B
			 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
			 */
			sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
			tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
			ofs = texture->header.w * tval + sval;
			if(texture->mask[ofs / 8] >> (7 - (ofs & 7)) & 1) {
				/* 1biteNX`ȂAeNZlɂ 1:tFCX̑OiF/0:tFCX̔wiF ŕ`܂B */
				*vptr = texture->buf[ofs / 8] >> (7 - (ofs & 7)) & 1 ? fore_color : back_color;
				*zptr = zval;
			}
		}
		/* J[Eyobt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}
static void
draw_scanline_2bit(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = CURRENT_TEXTURE(render);

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
			 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
			 * R[hTCYƂ̌ˍŁÂőË܂B
			 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
			 */
			sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
			tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
			ofs = texture->header.w * tval + sval;
			/* 2biteNX`ȂAeNZl̂܂ܕ`܂B */
			*vptr = texture->buf[ofs / 4] >> (3 - (ofs & 3)) * 2 & 3;
			*zptr = zval;
		}
		/* J[Eyobt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}
static void
draw_scanline_2bitmask(RENDER* render, const RENDVERTEX* v0, const RENDVERTEX* v1)
{
	/* DDAZbgAbvp}NB */
#define SETUP(i) \
	cur.i = v0->i; \
	if(v0->i <= v1->i) { \
		i.n = (v1->i - v0->i) % nx; \
		i.d = -nx; \
		i.u =  (v1->i - v0->i) / nx; \
		i.v =  1; \
	} else { \
		i.n = (v0->i - v1->i) % nx; \
		i.d = -nx; \
		i.u = -(v0->i - v1->i) / nx; \
		i.v = -1; \
	}

	/* DDAXebvsp}NB */
#define STEP(i) \
	cur.i += i.u; \
	i.d += i.n * 2; \
	if(i.d >= 0) { \
		cur.i += i.v; \
		i.d -= nx * 2; \
	}

	SURFACE* surface;	/* T[tFCX               */
	TEXTURE* texture;	/* eNX`                 */
	unsigned char* vptr;	/* J[obt@|C^     */
	unsigned char* zptr;	/* yobt@|C^         */
	int nx;			/* ݂̃XLC̒ */
	int n;			/* ݂̕ӂ̎csNZ   */
	RENDVERTEX cur;		/* ݒl                     */
	DDA z;			/* ylDDA                    */
	DDA s, t;		/* eNX`WDDA          */
	const RENDVERTEX* tmp;
	int ofs;
	unsigned char zval;
	unsigned char sval, tval;

	/* T[tFCXoĂ܂B */
	surface = render->surface;

	/* m :v0 -> v1:E Ƃ܂B */
	if(v0->x > v1->x) {
		tmp = v0;
		      v0 = v1;
		           v1 = tmp;
	}

	/* n_AhX߂܂B */
	ofs = surface->w * v0->y + v0->x;
	vptr = surface->vbuff + ofs;
	zptr = surface->zbuff + ofs;

	/* 厲DDAZbgAbvB */
	nx = v1->x - v0->x;
	if(nx == 0) return; /* _Ȃ`܂B */

	/* DDAZbgAbvB */
	SETUP(z);
	SETUP(s);
	SETUP(t);

	/* eNX`oĂ܂B */
	texture = CURRENT_TEXTURE(render);

	/* DDA[vB */
	n = nx;
	do {
		/* yrB */
		zval = (unsigned char)(cur.z >> 8);
		if(zval <= *zptr) {
			/* P/ECE͏Zx̂ŁAZ̗vEsv𔻒肵Ă珜Zsق܂}VłB
			 * eNX`WWrapȂꍇ̃XLC`R[hWJ΁Aɍł܂A
			 * R[hTCYƂ̌ˍŁÂőË܂B
			 * xr̎Qll: ɏZ(100%)->肵ďZ(150%)->R[hWJ(110%)
			 */
			sval = (unsigned char)(cur.s >> 8); if(sval >= texture->header.w) sval %= texture->header.w;
			tval = (unsigned char)(cur.t >> 8); if(tval >= texture->header.h) tval %= texture->header.h;
			ofs = texture->header.w * tval + sval;
			if(texture->mask[ofs / 8] >> (7 - (ofs & 7)) & 1) {
				/* 2biteNX`ȂAeNZl̂܂ܕ`܂B */
				*vptr = texture->buf[ofs / 4] >> (3 - (ofs & 3)) * 2 & 3;
				*zptr = zval;
			}
		}
		/* J[Eyobt@|C^i߂܂B */
		vptr++;
		zptr++;
		/* DDAXebvsB */
		STEP(z);
		STEP(s);
		STEP(t);
	} while(--n); /* E[sNZ͕`Ȃ̂ŁAvfNgłB */

#undef SETUP
#undef STEP
}
#endif /*OPT_DRAW_SCANLINE*/
