#define STRICT
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void usage();
void cmdline(int argc, char* argv[]);
void readbmp(FILE* fp);
void writeicon(FILE* fp)   { fprintf(stderr, "### NOT IMPLEMENTED YET\n"); exit(1); }
void writesprite(FILE* fp) { fprintf(stderr, "### NOT IMPLEMENTED YET\n"); exit(1); }
void writepbmp(FILE* fp, int bpp, int mask);

/****************************************************************************
 *
 ****************************************************************************/

/* /usr/piece/include/piece_bmp.h */
typedef struct _PBMP_FILEHEADER {
	unsigned long head;		/* + 0,4 VOl`'PBMP' */
	unsigned long fsize;		/* + 4,4 t@CTCY   */
	unsigned char bpp;		/* + 8,1 rbg[x       */
	unsigned char mask;		/* + 9,1 }XN[x       */
	short w;			/* +10,2                */
	short h;			/* +12,2              */
					/* +14,2 pfBO       */
	unsigned long buf_size;		/* +16,4 f[^TCY     */
					/* =20                    */
} PBMP_FILEHEADER;

/****************************************************************************
 *	
 ****************************************************************************/

#define NULDEV	"NUL"
FILE* msg = stdout;

#define TYPE_ICON	0
#define TYPE_SPRITE	1
#define TYPE_1BIT	2
#define TYPE_1BITMASK	3
#define TYPE_2BIT	4
#define TYPE_2BITMASK	5
char* out_type_desc[] = {
	"ICON",
	"SPRITE",
	"1BIT",
	"1BIT MASK",
	"2BIT",
	"2BIT MASK",
};
int out_type = -1;
char out_label[32];

char in_path[_MAX_PATH];
char out_path[_MAX_PATH];
char drive[_MAX_DRIVE];
char dir[_MAX_DIR];
char fname[_MAX_FNAME];
char ext[_MAX_EXT];

#define MAX_SIZE	1024
unsigned char in_data[MAX_SIZE][MAX_SIZE];
int in_width;
int in_height;
int out_width;
int out_height;

/****************************************************************************
 *	main
 ****************************************************************************/

int
main(int argc, char* argv[])
{
	FILE* fp;
	unsigned char d[] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 , 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };

	cmdline(argc, argv);

	fp = fopen(in_path, "rb");
	if(fp == NULL) {
		fprintf(stderr, "### %s open error\n", in_path);
		exit(1);
	}
	readbmp(fp);
	fclose(fp);


	switch(out_type) {
	case TYPE_ICON:
	case TYPE_SPRITE:
		fp = fopen(out_path, "wb");
		break;
	case TYPE_1BIT:
	case TYPE_1BITMASK:
	case TYPE_2BIT:
	case TYPE_2BITMASK:
		fp = fopen(out_path, "wt");
		break;
	default:
		abort();
	}
	if(fp == NULL) {
		fprintf(stderr, "### %s create error\n", out_path);
		exit(1);
	}
	switch(out_type) {
	case TYPE_ICON:
		writeicon(fp);
		break;
	case TYPE_SPRITE:
		writesprite(fp);
		break;
	case TYPE_1BIT:
		writepbmp(fp, 1, 0);
		break;
	case TYPE_1BITMASK:
		writepbmp(fp, 1, 1);
		break;
	case TYPE_2BIT:
		writepbmp(fp, 2, 0);
		break;
	case TYPE_2BITMASK:
		writepbmp(fp, 2, 1);
		break;
	default:
		abort();
	}
	fclose(fp);

	return 0;
}

/****************************************************************************
 *	usage
 ****************************************************************************/

void
usage()
{
	fprintf(stderr, "USAGE:\n");
	fprintf(stderr, "	dpbmpcnv [options] <filename.bmp>\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "TYPE OPTIONS:\n");
	fprintf(stderr, "	-ti  icon\n");
	fprintf(stderr, "	-ts  sprite\n");
	fprintf(stderr, "	-t1  1bit data\n");
	fprintf(stderr, "	-t1m 1bit data with mask\n");
	fprintf(stderr, "	-t2  2bit data\n");
	fprintf(stderr, "	-t2m 2bit data with mask\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "OTHER OPTIONS:\n");
	fprintf(stderr, "	-o<fname> filename\n");
	fprintf(stderr, "	-l<label> label name\n");
	fprintf(stderr, "	-q        quiet\n");
	exit(1);
}

/****************************************************************************
 *	cmdline
 ****************************************************************************/

void
cmdline(int argc, char* argv[])
{
	int i;
	char* opt;

	for(i = 1; i < argc; i++) {
		opt = argv[i];
		if(opt[0] != '-') {
			if(strlen(in_path) != 0) usage();
			strcpy(in_path, opt);
		} else {
			switch(opt[1]) {
			case 't':
				if(out_type != -1) usage();
				       if(strcmp(&opt[2], "i") == 0) {
					out_type = TYPE_ICON;
				} else if(strcmp(&opt[2], "s") == 0) {
					out_type = TYPE_SPRITE;
				} else if(strcmp(&opt[2], "1") == 0) {
					out_type = TYPE_1BIT;
				} else if(strcmp(&opt[2], "1m") == 0) {
					out_type = TYPE_1BITMASK;
				} else if(strcmp(&opt[2], "2") == 0) {
					out_type = TYPE_2BIT;
				} else if(strcmp(&opt[2], "2m") == 0) {
					out_type = TYPE_2BITMASK;
				} else {
					usage();
				}
				break;
			case 'o':
				if(strlen(out_path) != 0) usage();
				strcpy(out_path, &opt[2]);
				break;
			case 'l':
				if(strlen(out_label) != 0) usage();
				strcpy(out_label, &opt[2]);
				break;
			case 'q':
				msg = fopen(NULDEV, "wt");
				break;
			default:
				usage();
			}
		}
	}

	if(out_type == -1) usage();
	if(strlen(in_path) == 0) usage();
	if(strlen(out_path) == 0) {
		_splitpath(in_path, drive, dir, fname, ext);
		_makepath(out_path, drive, dir, fname, ".c");
	}
	if(strlen(out_label) == 0) {
		_splitpath(out_path, drive, dir, fname, ext);
		for(i = 0; fname[i] != '\0'; i++) {
			out_label[i] = toupper(fname[i]);
		}
	}

	fprintf(msg, "IN  FILE : %s\n", in_path);
	fprintf(msg, "OUT FILE : %s\n", out_path);
	fprintf(msg, "OUT_TYPE : %s\n", out_type_desc[out_type]);
	fprintf(msg, "OUT LABEL: %s\n", out_label);
}

/****************************************************************************
 *	readbmp
 ****************************************************************************/

void
readbmp(FILE* fp)
{
	int retval, pitch, y;
	BITMAPFILEHEADER bf;
	BITMAPINFOHEADER bi;

	/* t@Cwb_ǂ݂܂B */
	retval = fread(&bf, 1, sizeof(BITMAPFILEHEADER), fp);
	if(retval != sizeof(BITMAPFILEHEADER)) goto ERR;
	if(strncmp((char*)&bf.bfType, "BM", 2) != 0) goto ERR;

	/* rbg}bvwb_ǂ݂܂B */
	retval = fread(&bi, 1, sizeof(BITMAPINFOHEADER), fp);
	if(retval != sizeof(BITMAPINFOHEADER)) goto ERR;
	if(bi.biSize < sizeof(BITMAPINFOHEADER)) goto ERR;
	if(bi.biWidth > MAX_SIZE) goto ERR;
	if(bi.biHeight > MAX_SIZE) goto ERR;
	if(bi.biPlanes != 1) goto ERR;   /* 256FpbgBMP̂ݑΉ */
	if(bi.biBitCount != 8) goto ERR; /* 256FpbgBMP̂ݑΉ */
	if(bi.biCompression != BI_RGB) goto ERR;

	in_width = bi.biWidth;
	in_height = bi.biHeight;
	pitch = in_width + 3 & ~3; /* DIB̎dl */

	retval = fseek(fp, bf.bfOffBits, SEEK_SET);
	if(retval != 0) goto ERR;
	for(y = in_height - 1; y >= 0; y--) { /* {gAbv */
		retval = fread(&in_data[y][0], 1, pitch, fp);
		if(retval != pitch) goto ERR;
	}

	out_width = in_width + 7 & ~7; /* PBMP̎dl */
	out_height = in_height;

	return;
ERR:
	fprintf(stderr, "### readbmp error\n");
	exit(1);
}

/****************************************************************************
 *	writebmp
 ****************************************************************************/

void
writepbmp(FILE* fp, int bpp, int mask)
{
	PBMP_FILEHEADER hdr;
	int x, y, i;
	unsigned char in_byte, out_byte;

	/* Cz̊JnB */
	fprintf(fp, "const unsigned char %s[] = {\n", out_label);

	/* wb_o͂܂B */
	memset(&hdr, 0, sizeof hdr);
	memcpy(&hdr.head, "PMBP", 4); /* VOl` */
	         hdr.fsize = sizeof hdr + out_width * out_height / (8 / bpp); /* wb_ƃJ[̃TCY */
	if(mask) hdr.fsize +=             out_width * out_height /  8;        /* }XÑTCYǉ   */
	hdr.bpp  = bpp;
	hdr.mask = mask;
	hdr.w = out_width;
	hdr.h = out_height;
	         hdr.buf_size =  out_width * out_height / (8 / bpp); /* J[̃TCY       */
	if(mask) hdr.buf_size += out_width * out_height /  8;        /* }XÑTCYǉ */
	for(i = 0; i < sizeof hdr; i++) {
		fprintf(fp, "0x%02x,", ((unsigned char*)&hdr)[i]);
	}
	fprintf(fp, "\n");

	/* J[f[^o͂܂B */
	fprintf(fp, "/*buf*/\n");
	for(y = 0; y < out_height; y++) {
		for(x = 0; x < out_width; x += (8 / bpp)) {
			out_byte = 0;
			for(i = 0; i < (8 / bpp); i++) {
				out_byte <<= bpp;
				in_byte = in_data[y][x + i];
				if(in_byte < (1 << bpp)) { /* ̃J[0ƌȂ */
					out_byte |= in_byte;
				}
			}
			fprintf(fp, "0x%02x,", out_byte);
		}
		fprintf(fp, "\n");
	}

	/* }XNf[^o͂܂B */
	if(mask) {
		fprintf(fp, "/*mask*/\n");
		for(y = 0; y < out_height; y++) {
			for(x = 0; x < out_width; x += 8) {
				out_byte = 0;
				for(i = 0; i < 8; i++) {
					out_byte <<= 1;
					in_byte = in_data[y][x + i];
					if(in_byte < (1 << bpp)) {
						out_byte |= 1;
					}
				}
				fprintf(fp, "0x%02x,", out_byte);
			}
			fprintf(fp, "\n");
		}
	}

	/* Cz̏IB */
	fprintf(fp, "};\n");
}
